/**
 *  '$RCSfile: DBSAXNode.java,v $'
 *    Purpose: A Class that represents an XML node and its contents
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Matt Jones
 *    Release: @release@
 *
 *   '$Author: berkley $'
 *     '$Date: 2001/01/18 23:15:20 $'
 * '$Revision: 1.43 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import java.sql.*;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Enumeration;
import oracle.jdbc.driver.*;
import org.xml.sax.SAXException;

/** 
 * A Class that represents an XML node and its contents and
 * can write its own representation to a database connection
 */
public class DBSAXNode extends BasicNode {

  private Connection	conn;
  private DBSAXNode	parentNode;

  /** 
   * Construct a new node instance for DOCUMENT nodes
   *
   * @param conn the JDBC Connection to which all information is written
   */
  public DBSAXNode (Connection conn, String docid) throws SAXException {
    super();
    this.conn = conn;
    this.parentNode = null;
    writeChildNodeToDB("DOCUMENT", null, null, docid);
  }

  /** 
   * Construct a new node instance for ELEMENT nodes
   *
   * @param conn the JDBC Connection to which all information is written
   * @param tagname the name of the node
   * @param parentNode the parent node for this node being created
   */
  public DBSAXNode (Connection conn, String tagname, DBSAXNode parentNode, 
                    long rootnodeid, String docid, String doctype) 
                    throws SAXException {

    super(tagname);
    setParentID(parentNode.getNodeID());
    setRootNodeID(rootnodeid);
    setDocID(docid);
    setNodeIndex(parentNode.incChildNum());
    this.conn = conn;
    this.parentNode = parentNode;
    writeChildNodeToDB("ELEMENT", getTagName(), null, docid);
    //No writing XML Index from here. New Thread used instead.
    //updateNodeIndex(docid, doctype);
  }
    
  /** creates SQL code and inserts new node into DB connection */
  public void writeChildNodeToDB(String nodetype, String nodename,
                                 String data, String docid) 
                                 throws SAXException {
    try {
      PreparedStatement pstmt;
      if (nodetype == "DOCUMENT") {
        pstmt = conn.prepareStatement(
            "INSERT INTO xml_nodes " +
            "(nodeid, nodetype, nodename, docid, rootnodeid) " +
            "VALUES (?, ?, ?, ?, ?)");
        MetaCatUtil.debugMessage("INSERTING DOCNAME: " + nodename);
      } else {
        pstmt = conn.prepareStatement(
            "INSERT INTO xml_nodes " +
            "(nodeid, nodetype, nodename, docid, " +
            "rootnodeid, parentnodeid, nodedata, nodeindex) " +
            "VALUES (?, ?, ?, ?, ?, ?, ?, ?)");
      }

      // Bind the values to the query
      long nid = generateNodeID();
      pstmt.setLong(1, nid);
      pstmt.setString(2, nodetype);
      pstmt.setString(3, nodename);
      pstmt.setString(4, docid);
      if (nodetype == "DOCUMENT") {
        pstmt.setLong(5, nid);
      } else {
        if (nodetype == "ELEMENT") {
          pstmt.setLong(5, getRootNodeID());
          pstmt.setLong(6, getParentID());
          pstmt.setString(7, data);
          pstmt.setInt(8, getNodeIndex());
        } else {
          pstmt.setLong(5, getRootNodeID());
          pstmt.setLong(6, getNodeID());
          if ( nodetype == "TEXT" && getTagName().equals("meta_file_id") ) {
            pstmt.setString(7, getDocID());
          } else {
            pstmt.setString(7, data);
          }  
          pstmt.setInt(8, incChildNum());
        }
      }
      // Do the insertion
      pstmt.execute();
      pstmt.close();
      
      if (nodetype.equals("DOCUMENT")) {
        // Record the root node id that was generated from the database
        setRootNodeID(nid);
      }

      if (nodetype.equals("DOCUMENT") || nodetype.equals("ELEMENT")) {

        // Record the node id that was generated from the database
        setNodeID(nid);

        // Record the node type that was passed to the method
        setNodeType(nodetype);

      }

    } catch (SQLException e) {
      System.out.println("Error in DBSaxNode.writeChildNodeToDB");
      System.err.println("Error inserting node: (" + nodetype + ", " +
                                                     nodename + ", " + 
                                                     data + ")" );
      System.err.println(e.getMessage());
      throw new SAXException(e.getMessage());
    }
  }

  /** 
   * creates SQL code to put nodename for the document node 
   * into DB connection 
   */
  public void writeNodename(String nodename) throws SAXException {
      try {
        PreparedStatement pstmt;
        pstmt = conn.prepareStatement(
              "UPDATE xml_nodes set nodename = ? " +
              "WHERE nodeid = ?");

        // Bind the values to the query
        pstmt.setString(1, nodename);
        pstmt.setLong(2, getNodeID());
        // Do the insertion
        pstmt.execute();
        pstmt.close();
      } catch (SQLException e) {
        System.out.println("Error in DBSaxNode.writeNodeName: " + 
                           e.getMessage());
        throw new SAXException(e.getMessage());
      }
  }

  /** get next node id from DB connection */
  private long generateNodeID() throws SAXException {
      long nid=0;
      Statement stmt;
      try {
        stmt = conn.createStatement();
        stmt.execute("SELECT xml_nodes_id_seq.nextval FROM dual");
        ResultSet rs = stmt.getResultSet();
        boolean tableHasRows = rs.next();
        if (tableHasRows) {
          nid = rs.getLong(1);
        }
        stmt.close();
      } catch (SQLException e) {
        System.out.println("Error in DBSaxNode.generateNodeID: " + 
                            e.getMessage());
        throw new SAXException(e.getMessage());
      }

      return nid;
  }

  /** Add a new attribute to this node, or set its value */
  public void setAttribute(String attName, String attValue, String docid)
              throws SAXException {
    if (attName != null) {
      // Enter the attribute in the hash table
      super.setAttribute(attName, attValue);

      // And enter the attribute in the database
      writeChildNodeToDB("ATTRIBUTE", attName, attValue, docid);
    } else {
      System.err.println("Attribute name must not be null!");
      throw new SAXException("Attribute name must not be null!");
    }
  }

  /** 
   * NOT USED
   * Update the node index (xml_index) for this node by generating
   * test strings that represent all of the relative and absolute
   * paths through the XML tree from document root to this node
   */
  private void updateNodeIndex(String docid, String doctype) 
               throws SAXException
  {
    Hashtable pathlist = new Hashtable();
    boolean atStartingNode = true;
    boolean atRootDocumentNode = false;
    DBSAXNode nodePointer = this;
    StringBuffer currentPath = new StringBuffer();
    int counter = 0;

    // Create a Hashtable of all of the paths to reach this node
    // including absolute paths and relative paths
    while (!atRootDocumentNode) {
      if (atStartingNode) {
        currentPath.insert(0, nodePointer.getTagName());
        pathlist.put(currentPath.toString(), new Long(getNodeID()));
        counter++;
        atStartingNode = false;
      } else {
        currentPath.insert(0, "/");
        currentPath.insert(0, nodePointer.getTagName());
        pathlist.put(currentPath.toString(), new Long(getNodeID()));
        counter++;
      }

      // advance to the next parent node
      nodePointer = nodePointer.getParentNode();

      // If we're at the DOCUMENT node (root of DOM tree), add
      // the root "/" to make the absolute path
      if (nodePointer.getNodeType().equals("DOCUMENT")) {
        currentPath.insert(0, "/");
        pathlist.put(currentPath.toString(), new Long(getNodeID()));
        counter++;
        atRootDocumentNode = true;
      } 
    }

    try {
      // Create an insert statement to reuse for all of the path insertions
      PreparedStatement pstmt = conn.prepareStatement(
              "INSERT INTO xml_index (nodeid, path, docid, doctype, " + 
               "parentnodeid) " + 
              "VALUES (?, ?, ?, ?, ?)");
      ((OraclePreparedStatement)pstmt).setExecuteBatch(counter);
  
      pstmt.setString(3, docid);
      pstmt.setString(4, doctype);
      pstmt.setLong(5, getParentID());
      
      // Step through the hashtable and insert each of the path values
      Enumeration en = pathlist.keys();
      while (en.hasMoreElements()) {
        String path = (String)en.nextElement();
        Long nodeid = (Long)pathlist.get(path);
        pstmt.setLong(1, nodeid.longValue());
        pstmt.setString(2, path);
        pstmt.executeUpdate();
  
        //System.out.println(nodeid + " ==> " + path);
      }

      // Close the database statement
      pstmt.close();
    } catch (SQLException sqe) {
      System.err.println("SQL Exception while inserting path to index in " + 
                         "DBSAXNode.updateNodeIndex");
      System.err.println(sqe.getMessage());
      throw new SAXException(sqe.getMessage());
    }
  }

  /** 
   * USED FROM SEPARATE THREAD RUNNED from DBSAXHandler on endDocument()
   * Update the node index (xml_index) for this node by generating
   * test strings that represent all of the relative and absolute
   * paths through the XML tree from document root to this node
   */
  public void updateNodeIndex(Connection conn, String docid, String doctype) 
               throws SAXException
  {
    Hashtable pathlist = new Hashtable();
    boolean atStartingNode = true;
    boolean atRootDocumentNode = false;
    DBSAXNode nodePointer = this;
    StringBuffer currentPath = new StringBuffer();
    int counter = 0;

    // Create a Hashtable of all of the paths to reach this node
    // including absolute paths and relative paths
    while (!atRootDocumentNode) {
      if (atStartingNode) {
        currentPath.insert(0, nodePointer.getTagName());
        pathlist.put(currentPath.toString(), new Long(getNodeID()));
        counter++;
        atStartingNode = false;
      } else {
        currentPath.insert(0, "/");
        currentPath.insert(0, nodePointer.getTagName());
        pathlist.put(currentPath.toString(), new Long(getNodeID()));
        counter++;
      }

      // advance to the next parent node
      nodePointer = nodePointer.getParentNode();

      // If we're at the DOCUMENT node (root of DOM tree), add
      // the root "/" to make the absolute path
      if (nodePointer.getNodeType().equals("DOCUMENT")) {
        currentPath.insert(0, "/");
        pathlist.put(currentPath.toString(), new Long(getNodeID()));
        counter++;
        atRootDocumentNode = true;
      } 
    }

    try {
      // Create an insert statement to reuse for all of the path insertions
      PreparedStatement pstmt = conn.prepareStatement(
              "INSERT INTO xml_index (nodeid, path, docid, doctype, " + 
               "parentnodeid) " + 
              "VALUES (?, ?, ?, ?, ?)");
      ((OraclePreparedStatement)pstmt).setExecuteBatch(counter);
  
      pstmt.setString(3, docid);
      pstmt.setString(4, doctype);
      pstmt.setLong(5, getParentID());
      
      // Step through the hashtable and insert each of the path values
      Enumeration en = pathlist.keys();
      while (en.hasMoreElements()) {
        String path = (String)en.nextElement();
        Long nodeid = (Long)pathlist.get(path);
        pstmt.setLong(1, nodeid.longValue());
        pstmt.setString(2, path);
        pstmt.executeUpdate();
  
        //System.out.println(nodeid + " ==> " + path);
      }

      // Close the database statement
      pstmt.close();
    } catch (SQLException sqe) {
      System.err.println("SQL Exception while inserting path to index in " +
                         "DBSAXNode.updateNodeIndex");
      System.err.println(sqe.getMessage());
      throw new SAXException(sqe.getMessage());
    }
  }
 
  /** get the parent of this node */
  public DBSAXNode getParentNode() {
    return parentNode;
  }
}
